/** \file wiiheadtrack.cpp
 * 
 * Implementation of the Wiiheadtrack-class. Read the documentation for more 
 * information.
 * 
 * \author		Daniel Wunderlich (d.wunderlich@stud.uni-heidelberg.de)
 * \author		Christian Rohrbeck (c.rohrbec@stud.uni-heidelberg.de)
 * \version		1.0
 * \date		2010-09-06
 */

#include "wiiheadtrack.h"

Wiiheadtrack::Wiiheadtrack(WiihtMode _mode, unsigned int _noWiimotes) 
	: wiimoteCamNo(0)
{
	mode = _mode;
	connected = false;
	enabled = false;
	invY = 1;
	
	noWiimotes = _noWiimotes;
	foundWiimotes = 0;
	connectedWiimotes = 0;
	scanTime = 5;
	
	eyeDist[0] = 750.0;	eyeDist[1] = 600.0;
	posDist[0] = 6.0;	posDist[1] = 4.0;
	
	eyeL[0] = 0.0;	eyeL[1] = 0.0;
	eyeR[0] = 0.0;	eyeR[1] = 0.0;
	eyeC[0] = 0.0;	eyeC[1] = 0.0;
	eyeZ = 0;
	pos[0] = 0.0; pos[1] = 0.0; pos[2] = 0.0;
	posRel[0] = 0.0; posRel[1] = 0.0; posRel[2] = 0.0;
	
	theta = 0.0;
	phi = 0.0;
	radius = 0.0;
	
	if (mode == WIIHT_MODE_OBJECT) {
		sensitivity = 60.0;
	} else {
		sensitivity = 1.0;
	}
		
	wiimotes =  wiiuse_init(noWiimotes);
	std::cout << "[Wiiheadtrack] Created Wii Headtracker for " 
		<< noWiimotes << " Wiimotes!" << std::endl;	
	
	// initialize posDistPar
	computePosDistPar();
}

Wiiheadtrack::~Wiiheadtrack()
{
	wiiuse_cleanup(wiimotes, noWiimotes);
}

void Wiiheadtrack::connect()
{
	if (connected == false) {
		std::cout << "[Wiiheadtrack] Press 1 + 2 on the Wiimote!\n" 
			<< std::endl;
		foundWiimotes = wiiuse_find(wiimotes, noWiimotes, scanTime);
		if (!foundWiimotes) {
			std::cout <<"[Wiiheadtrack] Error: No Wiimote found!\n"
				<< std::endl;
		}
		connectedWiimotes = wiiuse_connect(wiimotes, noWiimotes); 
		if (connectedWiimotes) {
			connected = true;
			enabled = true;
			std::cout << "[Wiiheadtrack] Connected to " 
				<< connectedWiimotes << " Wiimotes !" << std::endl;
					
			// Set the position of the Wiimote below the screen
			wiiuse_set_ir_position(wiimotes[wiimoteCamNo], 
				WIIUSE_IR_BELOW);
			
			// Enable IR-tracking
			wiiuse_set_ir(wiimotes[wiimoteCamNo], 1);
			
			// Set IR-sensitivity
			wiiuse_set_ir_sensitivity(wiimotes[wiimoteCamNo], 1); 
					
		} else { 
			std::cerr << "[Wiiheadtrack] Error: Failed to connect to "
				<< "any Wiimote." << std::endl;
		}
	} else { 
		std::cout << "[Wiiheadtrack] Wiiheadtrack already connected."
			<< std::endl;
	}
}

void Wiiheadtrack::computePosDistPar()
{
	// compute the parameters m and d of the linear function 
	// m*x + d depending on posDist and eyeDist.
	posDistPar[0] = (posDist[1]-posDist[0])/(eyeDist[1]-eyeDist[0]);
	posDistPar[1] = posDist[0] - posDistPar[0]*eyeDist[0];
}

void Wiiheadtrack::computeEyePos()
{
	// poll the wiimote for new ir-position (right and left eye) 
	if (wiiuse_poll(wiimotes, noWiimotes)) {
		for (int i = 0; i < 2; i++) {
			if (wiimotes[wiimoteCamNo]->ir.dot[i].visible) {
				if (i == 0) {
					eyeR[0] = wiimotes[wiimoteCamNo]->ir.dot[i].x;
					eyeR[1] = wiimotes[wiimoteCamNo]->ir.dot[i].y;
				} else {
					eyeL[0] = wiimotes[wiimoteCamNo]->ir.dot[i].x;
					eyeL[1] = wiimotes[wiimoteCamNo]->ir.dot[i].y;
				}
			}
		}
		
		// compute the center of the eyes and translate them to the 
		// center
		eyeC[0] = ((eyeL[0] + eyeR[0]) * 0.5) - 512;
		eyeC[1] = ((eyeL[1] + eyeR[1]) * 0.5) - 384;
		
		// add offset from 0 to 0.5 * (eyeL[0] - eyeR[0]) from the 
		// center to the border, that eyeC is (theoretically) able to 
		// reach the boarder of the virtual screen of the Wiimote-camera
		eyeC[0] = eyeC[0] + eyeC[0]/512 * 0.5 * (eyeL[0] - eyeR[0]);
		eyeC[1] = eyeC[1] + eyeC[1]/384 * 0.5 * (eyeL[1] - eyeR[1]);
		
		if (wiimotes[wiimoteCamNo]->ir.z != 0) {
			eyeZ = wiimotes[wiimoteCamNo]->ir.z;
		}
	}
}

void Wiiheadtrack::computePlanePos()
{
	computeEyePos();
	
	pos[0] = eyeC[0]/512 * sensitivity;
	pos[1] = invY * eyeC[1]/384 * sensitivity;
	pos[2] = posDistPar[0] * eyeZ + posDistPar[1];
}

void Wiiheadtrack::computeSpherePos()
{
	computeEyePos();
	
	phi = sensitivity/180.0 * M_PI/512 * eyeC[0];
	theta = invY * sensitivity/180.0 * M_PI/384 * eyeC[1];
	
	float radius = posDistPar[0] * eyeZ + posDistPar[1];	
	
	pos[0] = radius * cos(theta) * sin(phi);
    pos[1] = radius * sin(theta);
    pos[2] = radius * cos(theta) * cos(phi);
}

void Wiiheadtrack::computePos()
{
  if (connected == true) {
	  float posTemp[3];
	  posTemp[0] = pos[0];
  	  posTemp[1] = pos[1];
	  posTemp[2] = pos[2];
	
	  if (mode == WIIHT_MODE_OBJECT && enabled == true) {
	  	computeSpherePos();
  	} else if (mode == WIIHT_MODE_ROOM && enabled == true) {
	    computePlanePos();
	  }
	
	  posRel[0] = posTemp[0] - pos[0];
  	  posRel[1] = posTemp[1] - pos[1];
	  posRel[2] = posTemp[2] - pos[2];
	}
}

void Wiiheadtrack::printStatus()
{
	std::cout 
		<< "--------------------Wiiheadtrack-status--------------------"
		<< std::endl
		<< "Connected Wiimotes: " << connectedWiimotes << ", mode = ";
		
	if (mode == WIIHT_MODE_OBJECT) {
		std::cout << "object-mode" << std::endl;
	} else if (mode == WIIHT_MODE_ROOM) {
		std::cout << "room-mode" << std::endl;
	}
		
	std::cout << "eyeDist = (" << eyeDist[0] << ", " 
		<< eyeDist[1] << "), "
		<< "posDist = (" << posDist[0] << ", " << posDist[1] << ")"
		<< std::endl
		<< "sensitivity = " << sensitivity
		<< std::endl
		<< "eyeL = (" << eyeL[0] << ", " << eyeL[1] << "), "	
		<< "eyeC = (" << eyeC[0] << ", " << eyeC[1] << "),"
		<< "eyeR = (" << eyeR[0] << ", " << eyeR[1] << ") "	
		<< std::endl
		<< "eyeZ = " << eyeZ
		<< std::endl;
	
	if (mode == WIIHT_MODE_OBJECT) {
		std::cout << "phi = " << phi/M_PI*180 << " deg, theta = " 
			<< theta/M_PI*180 << " deg" 
			<< std::endl
			<< "spherePos";
	} else if (mode == WIIHT_MODE_ROOM) {
		std::cout << "planePos";
	}
	std::cout << " = (" << pos[0] << ", " << pos[1] << ", " 
		<< pos[2] << ")" << std::endl;
	std::cout << "posRel = (" << posRel[0] << ", " << posRel[1] << ", " 
		<< posRel[2] << ")" << std::endl;
	std::cout 
		<< "-----------------------------------------------------------"
		<< std::endl;
}

void Wiiheadtrack::enable()
{
	enabled = true;
}

void Wiiheadtrack::disable()
{
	enabled = false;
}

void Wiiheadtrack::invertY()
{
	invY = -1;
}

bool Wiiheadtrack::isConnected()
{
	return connected;
}

bool Wiiheadtrack::isEnabled()
{
	return enabled;
}

float* Wiiheadtrack::getEyeDist()
{
	return eyeDist;
}

float* Wiiheadtrack::getPos()
{
	return pos;
}

float* Wiiheadtrack::getPosDist()
{
	return posDist;
}

float* Wiiheadtrack::getPosRel()
{
	return posRel;
}

float Wiiheadtrack::getSensitivity()
{
	return sensitivity;
}

void Wiiheadtrack::setCameraPos(WiihtCamPos pos)
{
	if (pos == WIIHT_BELOW) {
		wiiuse_set_ir_position(wiimotes[wiimoteCamNo], WIIUSE_IR_BELOW);
	} else {
		wiiuse_set_ir_position(wiimotes[wiimoteCamNo], WIIUSE_IR_ABOVE);
	}
}

void Wiiheadtrack::setEyeDist(float distCenter, float distNear)
{
	if (distCenter <= distNear || distCenter <= 0 || distNear <= 0) {
		std::cerr << "[Wiiheadtrack] Error: calling setEyeDist: " 
			<< "distCenter must be greater than distNear and both must "
			<< "be greater than 0. Variables not changed." 
			<< std::endl;
	} else {
		eyeDist[0] = distCenter;
		eyeDist[1] = distNear;
		computePosDistPar();
	}
}

void Wiiheadtrack::setPosDist(float distCenter, float distNear)
{
	if (distCenter <= distNear || distCenter <= 0 || distNear < 0) {
		std::cerr << "[Wiiheadtrack] Error: calling setPosDist: " 
			<< "distCenter must be greater than distNear, distCenter must "
			<< "be greater than 0 and distNear must be greater than or equal "
			<< "to 0. Variables not changed." 
			<< std::endl;
	} else {
		posDist[0] = distCenter;
		posDist[1] = distNear;
		computePosDistPar();
	}
}

void Wiiheadtrack::setScanTime(unsigned int time)
{
	scanTime = time;
}

void Wiiheadtrack::setSensitivity(float _sensitivity)
{
	if (_sensitivity > 0) {
		sensitivity = _sensitivity;
	} else {
		std::cerr << "[Wiiheadtrack] Error: calling setSensitivity: "
			<< "sensitivity must be greater than 0!" 
			<< std::endl;
	}
}
