/** \file wiiheadtrack.h
 * 
 * The header of the Wiiheadtrack-class. Read the documentation for more 
 * information.
 * 
 * \author		Daniel Wunderlich (d.wunderlich@stud.uni-heidelberg.de)
 * \author		Christian Rohrbeck (c.rohrbec@stud.uni-heidelberg.de)
 * \version		1.0
 * \date		2010-09-06
 */

#ifndef WIIHT_H
#define WIIHT_H

#include <iostream>
#include <math.h>
#include <GL/glut.h>
#include "wiiuse.h"

/** \brief	These values are used for the position of the virtual camera.
 * 	\see 	Wiiheadtrack::setCameraPos(WiihtCamPos)
 */
enum WiihtCamPos 
{
	WIIHT_BELOW,		/**< below the monitor*/
	WIIHT_ABOVE			/**< above the monitor*/
};

/** \brief	These values are used for the mode of Wiiheadtrack.
 * 	\see 	Wiiheadtrack::Wiiheadtrack(WiihtMode, unsigned int)
 */
enum WiihtMode 
{
	WIIHT_MODE_OBJECT, 	/**< object-mode*/
	WIIHT_MODE_ROOM		/**< room-mode*/
};

/**	\class Wiiheadtrack 
 * 	\brief	C++-class for using the Nintento Wii Remote (Wiimote) in combination 
 * 			with IR-LEDs attached to the user's head as a headtracking-system in 
 * 			graphic libraries.
 * 	
 * 	\author	Daniel Wunderlich (d.wunderlich@stud.uni-heidelberg.de)
 * 	\author Christian Rohrbeck (c.rohrbec@stud.uni-heidelberg.de)
 * 	\version 0.1
 * 	\date 2010-09-06
 */
class Wiiheadtrack
{
public:
	
/**	\brief Constructor of the Wiiheadtrack-class.
 * 	
 * 	\param 	noWiimotes	The number of Wiimotes you want to connect.
 * 	\param	mode		There are two modes, you can Wiiheadtrack use 
 * 						with: 
 * 		\li Object-mode: The virtual camera moves on the surface of a 
 * 			imaginary sphere, looking at the object in its center.
 * 		\li	Room-mode: The virtual camera moves on a plane, simulating 
 * 			the user watching through the monitor as a window 
 * 			into a "room" behind it.
 * 
 * 	\return A Wiiheadtrack-object.
 * 	\see	<a href="modes.html">The two modes</a>
 */ 
	Wiiheadtrack(WiihtMode mode, unsigned int noWiimotes);
	
	
/**	\brief Deconstructor of the Wiiheadtrack-class.
 */ 
	~Wiiheadtrack();


/**	\brief	Computes the current position (plane or spherical) of the 
 * 			user depending on the mode in 3 dimensions.
 * 
 * 			The results are are saved in a private array. For getting 
 * 			access to the results, use the getPos()- or getPosRel()-function.
 * 
 * 	\see	getPos()
 *  \see	getPosRel()
 * 	\see	<a href="position.html">Computation of the position</a>
 * 
 */
	void computePos();	
		
		
/**	\brief	Sets up the connection to the Wiimote.
 *
 *			After searching for available Wiimotes, the function tries 
 * 			to connect to them. It also sets the default position for
 * 			the "camera" (::WIIHT_BELOW) and enables the IR-tracking. You can 
 * 			set the time it tries connect with the help of 
 * 			setScanTime(unsigned int).
 */ 
	void connect();


/**	\brief	Enable Wiiheadtrack.
 *
 *			After connecting to a Wiimote, Wiiheadtrack is enabled by default.
 */ 
	void enable();
	

/**	\brief	Disable Wiiheadtrack.
 *
 *			Calling computePos() will not change the position anymore.
 */ 
	void disable();


/**	\brief	Because some systems invert the y-value of the current position, 
 * 			this function allows to invert the y-Axis position.
 */ 
	void invertY();
	
	
/**	\brief	Indicates, if Wiiheadtrack is connected to one or more Wiimote(s).
 */ 
	bool isConnected();
	
	
/**	\brief	Indicates, if Wiiheadtrack is enabled.
 */ 
	bool isEnabled();
	
	
/**	\brief	Prints several information about the current status of the 
 * 			Wiiheadtracker on cammand-line.
 * 
 * 			It prints: Wiiheadtrack::connected, Wiiheadtrack::mode, 
 * 			Wiiheadtrack::eyeDist, Wiiheadtrack::posDist, 
 * 			Wiiheadtrack::sensitivity, Wiiheadtrack::eyeL, Wiiheadtrack::eyeC, 
 * 			Wiiheadtrack::eyeR, Wiiheadtrack::eyeZ, Wiiheadtrack::phi, 
 * 			Wiiheadtrack::theta, Wiiheadtrack::pos, Wiiheadtrack::posRel.
 */
	void printStatus();
	

/**	\brief	This function is used for getting the eye-distance 
 * 			eyeDist[2].
 * 	
 * 	\return	A pointer to the 2-dimensional array saving the 
 * 			eye-distance.
 * 	\see	<a href="sensitivity.html">The Sensitivity</a>
 */ 
	float* getEyeDist();
	
		
/**	\brief	This function is used for getting the position-distance 
 * 			posDist[2].
 * 	
 * 	\return	A pointer to the 2-dimensional array saving the 
 * 			position-distance.
 * 	\see	<a href="sensitivity.html">The Sensitivity</a>
 */ 
	float* getPosDist();


/**	\brief	This function is used for accessing the current position 
 * 			(plane or spherical) of the user depending on the mode.
 * 
 * 			The position previously must be computed by computePos().
 * 	
 * 	\return	A pointer to the 3-dimensional array saving the current 
 * 			position.
 * 	\see	computePos()
 * 	\see	<a href="position.html">Computation of the position</a>
 */
	float* getPos();
	
	
/**	\brief	This function is used for accessing the current relative 
 * 			position (plane or spherical) of the user depending on the 
 * 			mode.
 * 
 * 			The position previously must be computed by computePos().
 * 
 * 	\return	A pointer to the 3-dimensional array saving the current 
 * 			relative position.
 * 	\see	computePos()
 * 	\see	<a href="position.html">Computation of the position</a>
 */
	float* getPosRel();
	

/**	\brief	Get the sensitivity of Wiiheadtrack.
 *  	
 * 	\return	The (x/y-)sensitivity of the Wiiheadtrack.
 * 	\see	setSensitivity(float)
 * 	\see	<a href="sensitivity.html">Sensitivity</a>
 */ 
	float getSensitivity();
	

/**	\brief	Set the position of the camera -- above or below the monitor.
 * 	
 * 	\param	pos	A ::WiihtCamPos-value.
 */
	void setCameraPos(WiihtCamPos pos);


/**	\brief	Set the "normal" an the smallest distance between the 
 * 			eyes/user an the camera.
 * 
 * 			The distances are used for configuring the sensitivity in 
 * 			z-direction.
 * 	
 * 	\param 	distCenter	The "normal" distance between the eye/user and the 
 * 						camera. Must be greater than distNear and greater than 
 * 						0.
 * 	\param	distNear	The smallest distance between the eye/user and the 
 * 						camera. Must be greater than 0.
 * 	\see	setPosDist(float, float)
 * 	\see	<a href="sensitivity.html">The Sensitivity</a>
 */ 
	void setEyeDist(float distCenter, float distNear);


/**	\brief	Set the time (seconds) the program tries to connect to the 
 * 			wiimote. (Default: 5 seconds)
 * 	
 * 	\param 	time	The time in seconds.
 */ 
	void setScanTime(unsigned int time);
	
	
/**	\brief	Set the sensitivity of Wiiheadtrack. (Default: 60.0 (object-mode), 
 * 			1.0 (room-mode))
 * 
 * 			The sensitivity describes, how intense the camera moves
 * 			depending on the head-movement.
 * 	
 * 	\param 	sensitivity	The new sensitivity. Must be greater 0.
 */ 
	void setSensitivity(float sensitivity);


/**	\brief	Set the "normal" an the smallest distance between the 
 * 			(virtual) position an the (virtual) camera-center.
 * 
 * 			The distances are used for configuring the sensitivity in 
 * 			z-direction. 
 * 	
 * 	\param 	distCenter	The "normal" distance between the camera and 
 * 						the center. Must be greater than distNear and 
 * 						greater than 0.
 * 	\param	distNear	The smallest distance between the camera and 
 * 						the center. Must be greater than or equal to 0
 * 	\see 	setEyeDist(float, float);
 * 	\see	<a href="sensitivity.html">The Sensitivity</a>
 */ 
	void setPosDist(float distCenter, float distNear);
	
	
private:

	WiihtMode mode;						/**< The mode you want to use the Wiiheadtrack. \see Wiiheadtrack(WiihtMode, unsigned int)*/
	wiimote** wiimotes;					/**< Array of the Wiimotes.*/
	bool connected;						/**< Indicates, if Wiiheadtrack is connected to one or more Wiimote(s)*/
	bool enabled;                       /**< Indicates, if Wiiheadtracking is enabled.*/
	int invY;							/**< Some systems invert the y-Value. This variable helps to fix this error. Inverted=-1, not inverted=1. \see invertY() */
	
	unsigned int noWiimotes;			/**< Number of Wiimotes you want to connect. Default: 1*/
	unsigned int foundWiimotes;			/**< Number of Wiimotes found by scanning in the connect()-function.*/
	unsigned int connectedWiimotes;		/**< Number of Wiimotes connected with in the connect()-function.*/
	unsigned int scanTime;				/**< Time (seconds) Wiiheadtrack searches for Wiimotes in the connect()-function.*/
	
	float eyeDist[2];					/**< Distance between the eye/user and the camera: "normal" distance (eyeDist[0]), nearest position to the camera (eyeDist[1]). \see setEyeDist(float, float) \see <a href="sensitivity.html">Sensitivity</a>*/
	float posDist[2];					/**< Distance between the "normal" virtual position an the nearest virtual position. \see setPosDist(float, float) \see <a href="sensitivity.html">Sensitivity</a>*/
	float posDistPar[2];				/**< Parameters \f$m\f$ and \f$d\f$ of the linear function \f$f(x) = m \cdot x + d\f$ for mapping the eye-distance to the position-distance. \see computePosDistPar() \see <a href="sensitivity.html">Sensitivity</a>*/

	float eyeL[2];						/**< 2-dimensional position of the user's left eye on the imaginary screen seen by the camera.*/
	float eyeR[2];						/**< 2-dimensional position of the user's right eye on the imaginary screen seen by the camera.*/
	float eyeC[2];						/**< 2-dimensional position of the point between the user's eyes on the imaginary screen seen by the camera.*/
	float eyeZ;							/**< Distance between the Wiimote-camera and the user's eyes.*/
	float pos[3];						/**< Position computed by computePos(), depending on the mode.*/
	float posRel[3];					/**< Change of pos[3] since the last call of computePos() (relative Position).*/
	
	const int wiimoteCamNo;				/**< Indicates, which of the connected Wiimotes is used as the camera. Default: 1.*/
	float sensitivity;					/**< A factor which declares, how intense the movement (left/right, up/down) of the user moves the virtual camera. \see setSensitivity(float) \see <a href="sensitivity.html">Sensitivity</a>*/
	float phi, theta, radius;			/**< Values for calculating the spherical coordinates of the virtual camera/position.*/
	
	
/**	\brief	Detects the position of the user on the imaginary screen 
 * 			seen by the "camera".
 * 
 * 			After detecting the position of the left (eyeL) and the 
 * 			right (eyeR) eye, the function calculates the point 
 * 			between the eyes. It also detects the z-value of the 
 * 			position of the users eyes seen by the Wiimote-camera.
 * 
 * \see		<a href="position.html">Computation of the position</a>
 */
	void computeEyePos();
	
	
/**	\brief	Transforms the position of the user detected by 
 * 			computeEyePos() to a virtual position on a plane, 
 * 			regarding the sensitivity. Used in room-mode.
 * 
 * 	\see	<a href="position.html">Computation of the position</a>
 */
	void computePlanePos();
	
	
/** \brief	Computes the parameters \f$m\f$ and \f$d\f$ of the linear 
 * 			function \f$f(x) = m \cdot x + d\f$ for mapping the 
 * 			eye-distance to the position-distance.
 * 	
 *	\see	<a href="sensitivity.html">The Sensitivity</a>
 * 
 */
	void computePosDistPar();
	
	
/**	\brief	Maps the position of the user detected by 
 * 			computeEyePos() to a virtual position on a sphere, 
 * 			regarding the sensitivity. Used in object-mode.
 * 
 * 	\see	<a href="position.html">Computation of the position</a>
 */
	void computeSpherePos();
};

#endif /* WIIHT_H */

